#include "client_socket.h"
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

namespace sxx_socket {

int client_socket::create_socket() {
	int fd = socket(AF_INET, SOCK_STREAM, 0);

	if (fd == -1)
		throw_creation_exception();

	return fd;
}

void client_socket::init_socket() {
	int yes = 1;

	setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &yes, sizeof(yes));
	setsockopt(fd, SOL_SOCKET, SO_NOSIGPIPE, &yes, sizeof(yes));
}

bool client_socket::is_space(char x) {
	return isspace(x);
}

void client_socket::throw_receive_timeout() {
	socket_receive_timeout e;
	throw e;
}

void client_socket::throw_transfer_timeout() {
	socket_transfer_timeout e;
	throw e;
}

void client_socket::throw_creation_exception() {
	socket_creation_exception e;
	throw e;
}

void client_socket::throw_connection_exception() {
	socket_connection_exception e;
	throw e;
}

client_socket::client_socket(int fd) {
	this->fd = (fd != -1)? fd: create_socket();
	init_socket();
}

client_socket::client_socket(const client_socket& other) {
	fd = dup(other.fd);
	if (fd == -1)
		throw_creation_exception();
}

client_socket::~client_socket() {
	close(fd);
}

void client_socket::connect_to(const string& host, const string& port) {
	bool connected = false;
	struct addrinfo* list = NULL;
	struct addrinfo* ptr = NULL;
	struct addrinfo hint;

	memset(&hint, 0, sizeof(hint));
	hint.ai_family = AF_INET;
	hint.ai_socktype = SOCK_STREAM;

	getaddrinfo(host.c_str(), port.c_str(), &hint, &list);

	for (ptr = list; ptr; ptr = ptr->ai_next) {
		if (!connect(fd, ptr->ai_addr, sizeof(struct sockaddr_in))) {
			connected = true;
			break;
		}
	}

	if (list)
		freeaddrinfo(list);

	if (!connected)
		throw_connection_exception();
}

char client_socket::get_char() {
	char temp = 0;
	
	if (read(fd, &temp, 1) != 1)
		throw_receive_timeout();

	return temp;
}

string client_socket::get_word() {
	char temp = 0;
	string result;

	// Skip the leading spaces.
	while (true) {
		if (read(fd, &temp, 1) != 1)
			throw_receive_timeout();

		if (!is_space(temp)) {
			result += temp;
			break;
		}
	}

	// Concat the word.
	while (true) {
		if (read(fd, &temp, 1) != 1)
			throw_receive_timeout();

		if (!is_space(temp))
			result += temp;
		else
			break;
	}

	return result;
}

string client_socket::get_string(size_t size) {
	char temp = 0;
	string result(size, 0);
	string::iterator itr = result.begin();

	while (size--) {
		if (read(fd, &temp, 1) != 1)
			throw_receive_timeout();
		*itr = temp;
		++itr;
	}

	return result;
}

string client_socket::get_line() {
	char temp = 0;
	string result;

	while (temp != '\n' && read(fd, &temp, 1) == 1)
		result += temp;

	if (result == "")
		throw_receive_timeout();

	// Chop off the excess \n, or \r\n.
	if (result.length() >= 2 && result[result.length() - 2] == '\r')
		return result.substr(0, result.length() - 2);
	else if (result[result.length()] == '\n')
		return result.substr(0, result.length() - 1);
	else
		return result;
}

void client_socket::put_char(const char& content) {
	if (write(fd, &content, 1) != 1)
		throw_transfer_timeout();
}

void client_socket::put_string(const string& content) {
	if (write(fd, content.c_str(), content.length())
	    != (int)content.length())
		throw_transfer_timeout();
}

void client_socket::put_word(const string& content) {
	put_string(content);
	put_string(" ");
}

void client_socket::put_line(const string& content) {
	put_string(content);
	put_line();
}

void client_socket::put_line() {
	static const char* end_line = "\r\n";
	static const size_t end_line_length = 2;

	if (write(fd, end_line, end_line_length)
	    != (int)end_line_length)
		throw_transfer_timeout();
}

void client_socket::set_receive_timeout(int seconds) {
	struct timeval timeout;
	timeout.tv_sec = seconds;
	timeout.tv_usec = 0;

	setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));
}

void client_socket::set_transfer_timeout(int seconds) {
	struct timeval timeout;
	timeout.tv_sec = seconds;
	timeout.tv_usec = 0;

	setsockopt(fd, SOL_SOCKET, SO_SNDTIMEO, &timeout, sizeof(timeout));
}

const client_socket& client_socket::operator = (const client_socket& other) {
	if (this == &other)
		return *this;

	close(fd);
	fd = dup(other.fd);

	if (fd == -1)
		throw_creation_exception();

	return *this;
}

const char* socket_failure::what() const throw() {
	return "Socket failure.";
}

const char* socket_receive_timeout::what() const throw() {
	return "Receive timeout.";
}

const char* socket_transfer_timeout::what() const throw() {
	return "Transfer timeout.";
}

const char* socket_creation_exception::what() const throw() {
	return "Unable to create socket.";
}

const char* socket_connection_exception::what() const throw() {
	return "Unable to connect.";
}

} // end of namespace sxx_socket.

