#include "connection_manager.h"
#include <iostream>
#include <algorithm>
using namespace std;

namespace sxx_manager {

connection_info::connection_info(const client_socket& client,
                                 const string& host, const string& port) {
	this->client = client;
	this->host = host;
	this->port = port;
}

connection_info::connection_info(const connection_info& value) {
	this->client = value.client;
	this->host = value.host;
	this->port = value.port;
}

void connection_manager::signal_handler(int signal) {
	// cerr << "Signal " << signal << " caught." << endl;
	get_instance().close_connection(pthread_self());
	pthread_exit(0);
}

void* connection_manager::handler_wrapper(void* argv) {
	connection_manager& instance = get_instance();

	if (!instance.handler)
		return 0;

	instance.handler(*((connection_info*)argv));
	instance.close_connection(pthread_self());
	return 0;
}

void* connection_manager::thread_collector(void* argv) {
	void* dummy_ptr = 0;
	map<pthread_t, connection_info*>::const_iterator itr;
	connection_manager& instance = get_instance();

	while (true) {
		sleep(1);
		pthread_mutex_lock(&(instance.collect_threads_lock));

		for (itr = instance.collect_threads.begin();
		     itr != instance.collect_threads.end();
		     ++itr) {
			pthread_join(itr->first, &dummy_ptr);
			delete itr->second;
			// cout << "Thread " << itr->first << " recycled." << endl;
		}
		instance.collect_threads.clear();

		pthread_mutex_unlock(&(instance.collect_threads_lock));
	}

	return 0;
}

connection_manager::connection_manager() {
	pthread_t thread; // dummy variable.
	pthread_mutex_init(&running_threads_lock, NULL);
	pthread_mutex_init(&collect_threads_lock, NULL);

	this->handler = 0;
	this->thread_limit = 100;

	pthread_create(&thread, NULL, thread_collector, NULL);
}

connection_manager& connection_manager::get_instance() {
	static connection_manager instance;
	return instance;
}

void connection_manager::set_handler(void (*handler)(connection_info& info)) {
	this->handler = handler;
}

void connection_manager::set_limit(size_t count) {
	thread_limit = count;
}

void connection_manager::new_connection(connection_info& info) {
	connection_info* info_copy_ptr;
	pthread_t new_thread;

	// cout << "New connection from " << info.host << ":"
	//      << info.port << "... " << flush;

	info_copy_ptr = new connection_info(info);
	pthread_create(&new_thread, NULL, handler_wrapper, info_copy_ptr);

	while (true) {
		pthread_mutex_lock(&running_threads_lock);
		if (running_threads.size() < thread_limit) {
			running_threads[new_thread] = info_copy_ptr;
			pthread_mutex_unlock(&running_threads_lock);
			break;
		}
		pthread_mutex_unlock(&running_threads_lock);
		sleep(1);
	}

	// cout << "Thread " << new_thread << " allocated." << endl;
}

void connection_manager::close_connection(pthread_t thread) {
	connection_info* info_ptr;

	// cout << "Thread " << thread << " is pending to be recycled." << endl;

	pthread_mutex_lock(&running_threads_lock);
	info_ptr = running_threads[thread];
	running_threads.erase(thread);
	pthread_mutex_unlock(&running_threads_lock);

	pthread_mutex_lock(&collect_threads_lock);
	collect_threads[thread] = info_ptr;
	pthread_mutex_unlock(&collect_threads_lock);
}

void connection_manager::close_all_connections() {
	map<pthread_t, connection_info*>::const_iterator itr;
	
	pthread_mutex_lock(&running_threads_lock);
	thread_limit = 0;
	pthread_mutex_unlock(&running_threads_lock);
	
	while (1) {
		pthread_mutex_lock(&running_threads_lock);
		if (running_threads.empty())
			break;
		pthread_mutex_unlock(&running_threads_lock);
		sleep(1);
	}
}

} // end of namespace sxx_manager.

