#include "file_manager.h"
using namespace sxx_file;

#define file_count 200

namespace sxx_file {

file_manager* file_manager::me;

file_t::file_t() {

}

file_t::file_t(const file_t& other) {
	pthread_mutex_init(&write_mutex, NULL);
	pthread_mutex_init(&read_mutex, NULL);
	filename = other.filename;
	last_access = other.last_access;
	content = other.content;
}

file_t::file_t(const string& filename, bool no_read) {
	pthread_mutex_init(&write_mutex, NULL);
	pthread_mutex_init(&read_mutex, NULL);
	last_access = time(NULL);
	this->filename = filename;

	if (no_read)
		return;

	ifstream input(filename.c_str());
	if (!input) {
		access_denied err;
		throw err;
	}

	char buffer[1024];
	size_t bytes_read = 0;
	do {
		input.read(buffer, 1024);
		bytes_read = input.gcount();
		content += string(buffer, bytes_read);
	} while (bytes_read > 0);
}

file_t::~file_t() {
	if (filename.empty())
		return;
	
	pthread_mutex_destroy(&write_mutex);
	pthread_mutex_destroy(&read_mutex);


	ofstream output(filename.c_str());

	if (!output) {
		access_denied err;
		throw err;
	}

	output.write(content.c_str(), content.length());
}

void file_t::write_lock() {
	pthread_mutex_lock(&write_mutex);
	pthread_mutex_lock(&read_mutex);
	last_access = time(NULL);
}

void file_t::read_lock() {
	pthread_mutex_lock(&write_mutex);
	pthread_mutex_lock(&read_mutex);
	pthread_mutex_unlock(&write_mutex);
	last_access = time(NULL);
}

void file_t::release_write_lock() {
	pthread_mutex_unlock(&write_mutex);
	pthread_mutex_unlock(&read_mutex);
	last_access = time(NULL);
}

void file_t::release_read_lock() {
	pthread_mutex_unlock(&read_mutex);
	last_access = time(NULL);
}

file_t& file_manager::locate_file(string filename, bool no_read) {
	map<string, file_t>::iterator itr;

	pthread_mutex_lock(&master_lock);

	// Try to locate the file.
	itr = cache.find(filename);
	if (itr != cache.end()) {
		pthread_mutex_unlock(&master_lock);
		return itr->second;
	}

	// If the file does not exist in cache, check if there are too many
	// cache. If yes, swap some out.
	if (cache.size() >= file_count && cache.begin() != cache.end()) {
		map<string, file_t>::iterator to_delete;
		time_t oldest = 0;

		to_delete = cache.begin();
		oldest = cache.begin()->second.last_access;
		for (itr = cache.begin(); itr != cache.end(); ++itr) {
			if (itr->second.last_access < oldest) {
				to_delete = itr;
				oldest = itr->second.last_access;
			}
		}

		to_delete->second.write_lock();
		cache.erase(to_delete);
	}

	// Now, load a new ...
	try {
		cache[filename] = file_t(filename, no_read);
	} catch (...) {
		cache.erase(filename);
		pthread_mutex_unlock(&master_lock);
		throw;
	}
	
	// Return the reference.
	pthread_mutex_unlock(&master_lock);
	return cache[filename];
}

file_manager::file_manager() {
	pthread_mutex_init(&master_lock, NULL);
}

file_manager::~file_manager() {
	map<string, file_t>::iterator i;
	pthread_mutex_lock(&master_lock);
	cache.clear();
	pthread_mutex_destroy(&master_lock);
}

file_manager& file_manager::instance() {
	static bool init = false;

	if (!init) {
		me = new file_manager();
		init = true;
	}

	return *me;
}

void file_manager::destroy() {
	delete me;
	me = 0;
}

string file_manager::read_file(const string& filename) {
	string result;
	file_t& object = locate_file(filename, false);
	object.read_lock();
	result = object.content;
	object.release_read_lock();
	return result;
}

void file_manager::write_file(const string& filename, const string& content) {
	file_t& object = locate_file(filename, true);
	object.write_lock();
	object.content = content;
	object.release_write_lock();
}

const char* file_manager_failure::what() const throw() {
	return "Unable to operate on file.";
}

const char* access_denied::what() const throw() {
	return "Access Denied.";
}

} // end of namespace sxx_file

