#include "response_generator.h"

namespace sxx_generator {

const char* response_generator::status_description(int status) {
	switch (status) {
	case 200:
		return "OK";
	case 400:
		return "Bad Request";
	case 401:
		return "Unauthorized";
	case 403:
		return "Forbidden";
	case 404:
		return "Not Found";
	case 411:
		return "Length Required";
	case 500:
		return "Internal Server Error";
	case 301:
		return "Moved Permanently";
	case 302:
		return "Moved Temporarily";
	case 303:
		return "See Other";
	case 304:
		return "Not Modified";
	case 305:
		return "Use Proxy";
	default:
		return "Programmer Defined Response";
	}
}

string response_generator::normalize(const string& str) {
	string result(str);
	string::iterator itr;
	char last_char = 0;

	for (itr = result.begin(); itr != result.end(); ++itr) {
		if (last_char && last_char != '-') {
			if (*itr >= 'A' && *itr <= 'Z') {
				*itr -= 'A';
				*itr += 'a';
			}
		} else {
			if (*itr >= 'a' && *itr <= 'z') {
				*itr -= 'a';
				*itr += 'A';
			}
		}
		last_char = *itr;
	}

	return result;
}

string response_generator::hex_encode_char(unsigned char x) {
	static char* hex_chars = "0123456789ABCDEF";

	int upper = x >> 4;
	int lower = x - ((x >> 4) << 4);
	string result;

	result += '%';
	result += hex_chars[upper];
	result += hex_chars[lower];
	return result;
}

string response_generator::hex_encode(const string& str) {
	string result;
	string::const_iterator itr;

	for (itr = str.begin(); itr != str.end(); ++itr) {
		if (*itr == ' ')
			result += '+';
		else if (*itr == '%' || *itr == ';' || *itr == '+'
		         || *itr > 126 || *itr < 32)
			result += hex_encode_char(*itr);
		else
			result += *itr;
	}

	return result;
}

void response_generator::initial(const response& response) {
	output.put_word("HTTP/1.1");
	output.put_word((string)(var)response.status);
	output.put_line(status_description(response.status));
}

void response_generator::added_headers(const response& response) {
	map<string, string>::const_iterator itr;
	for (itr = response.headers.begin();
	     itr != response.headers.end();
	     ++itr) {
		output.put_word(normalize(itr->first) + ":");
		output.put_line(itr->second);
	}
}

void response_generator::cookie_headers(const response& response) {
	map<string, var>::const_iterator itr;
	string epoxy;
	string one_year;
	char buffer[64];
	time_t epoxy_t = 0;
	time_t one_year_t = time(NULL) + 86400 * 365;

	strftime(buffer, 64, "%a, %d %b %Y %H:%M:%S %Z", gmtime(&epoxy_t));
	epoxy = buffer;

	strftime(buffer, 64, "%a, %d %b %Y %H:%M:%S %Z", gmtime(&one_year_t));
	one_year = buffer;

	for (itr = response.expire_cookie.begin();
	     itr != response.expire_cookie.end();
	     ++itr) {
		output.put_string("Set-Cookie: ");
		output.put_string(hex_encode(itr->first));
		output.put_string("=" + hex_encode((string)itr->second));
		output.put_string("; expires=" + epoxy);
		output.put_string("; path=/");
		output.put_line();
	}

	for (itr = response.session_cookie.begin();
	     itr != response.session_cookie.end();
	     ++itr) {
		output.put_string("Set-Cookie: ");
		output.put_string(hex_encode(itr->first));
		output.put_string("=" + hex_encode((string)itr->second));
		output.put_string("; path=/");
		output.put_line();
	}

	for (itr = response.persistent_cookie.begin();
	     itr != response.persistent_cookie.end();
	     ++itr) {
		output.put_string("Set-Cookie: ");
		output.put_string(hex_encode(itr->first));
		output.put_string("=" + hex_encode((string)itr->second));
		output.put_string("; expires=" + one_year);
		output.put_string("; path=/");
		output.put_line();
	}
}

void response_generator::misc_headers(const response& response) {
	output.put_line("Server: S++");

	if (response.connection_close)
		output.put_line("Connection: close");

	if (!response.location.empty()) {
		output.put_word("Location:");
		output.put_line(hex_encode(response.location));
	}
}

void response_generator::content_headers(const response& response) {
	output.put_word("Content-Type:");
	output.put_line(response.content_type);

	output.put_word("Content-Length:");
	output.put_line((string)(var)(response.content.str().size()));
}

void response_generator::content(const response& response) {
	output.put_string(response.content.str());
}

response_generator::response_generator(const client_socket& output) {
	this->output = output;
}

void response_generator::put_response(const response& response) {
	initial(response);
	added_headers(response);
	cookie_headers(response);
	misc_headers(response);
	content_headers(response);
	output.put_line();
	content(response);

	if (response.connection_close) {
		response_generator_close_exception e;
		throw e;
	}
}

const char* response_generator_failure::what() const throw() {
	return "Response generator failure.";
}

const char* response_generator_close_exception::what() const throw() {
	return "Connection should close.";
}

} // end of namespace sxx_generator.

