#include "server_socket.h"
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

namespace sxx_socket {

int server_socket::create_socket() {
	int fd = socket(AF_INET, SOCK_STREAM, 0);

	if (fd == -1)
		throw_creation_exception();

	return fd;
}

void server_socket::init_socket() {
	int yes = 1;

	setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &yes, sizeof(yes));
	setsockopt(fd, SOL_SOCKET, SO_NOSIGPIPE, &yes, sizeof(yes));
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
}

void server_socket::bind_socket(const struct sockaddr_in* address) {
	if (bind(fd,
	         (const struct sockaddr*) address,
	         sizeof(struct sockaddr_in)))
		throw_binding_exception();
}

void server_socket::listen_socket(int backlog) {
	if (listen(fd, backlog))
		throw_listen_exception();
}

void server_socket::throw_accept_timeout() {
	socket_accept_timeout e;
	throw e;
}

void server_socket::throw_binding_exception() {
	socket_binding_exception e;
	throw e;
}

void server_socket::throw_listen_exception() {
	socket_listen_exception e;
	throw e;
}

void server_socket::throw_creation_exception() {
	socket_creation_exception e;
	throw e;
}

server_socket& server_socket::operator = (const server_socket& other) {
	// Do nothing.
	return *this;
}

server_socket::server_socket(const string& host, const string& port,
                             int backlog) {
	struct addrinfo* list = NULL;
	struct addrinfo hint;

	memset(&hint, 0, sizeof(hint));
	hint.ai_family = AF_INET;
	hint.ai_socktype = SOCK_STREAM;

	getaddrinfo(host.c_str(), port.c_str(), &hint, &list);

	if (!list)
		throw_binding_exception();

	fd = create_socket();
	init_socket();

	try {
		bind_socket((struct sockaddr_in*)list->ai_addr);
		listen_socket(backlog);
	} catch (const exception& e) {
		freeaddrinfo(list);
		close(fd);
		throw;
	}

	freeaddrinfo(list);
}

client_socket server_socket::accept_client() {
	string host;
	string port;
	return accept_client(host, port);
}

client_socket server_socket::accept_client(string& host, string& port) {
	int client_fd = -1;
	struct sockaddr_in remote_addr;
	socklen_t remote_addr_length = sizeof(remote_addr);
	char buffer[16];

	client_fd = accept(fd,
	                   (struct sockaddr*)&remote_addr,
	                   &remote_addr_length);

	if (client_fd == -1)
		throw_accept_timeout();

	client_socket result(client_fd);
	result.set_receive_timeout(receive_timeout);
	result.set_transfer_timeout(transfer_timeout);

	host = inet_ntoa(remote_addr.sin_addr);
	sprintf(buffer, "%d", ntohs(remote_addr.sin_port));
	port = buffer;

	return result;
}

void server_socket::set_receive_timeout(int seconds) {
	receive_timeout = seconds;
}

void server_socket::set_transfer_timeout(int seconds) {
	transfer_timeout = seconds;
}

void server_socket::set_accept_timeout(int seconds) {
	struct timeval timeout;
	timeout.tv_sec = seconds;
	timeout.tv_usec = 0;

	setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));
}

const char* socket_accept_timeout::what() const throw() {
	return "Accept timeout.";
}

const char* socket_binding_exception::what() const throw() {
	return "Unable to bind.";
}

const char* socket_listen_exception::what() const throw() {
	return "Unable to listen.";
}

} // end of namespace sxx_socket.

