#include "var.h"
#include <sstream>
using std::ostringstream;
using std::istringstream;

namespace sxx_var {

void var::free_memory() {
	switch (current_type) {
	case s:
		delete s_ptr;
		current_type = e;
		break;
	case v:
		delete v_ptr;
		current_type = e;
		break;
	case m:
		delete m_ptr;
		current_type = e;
		break;
	default:
		// Do nothing as - no dynamically allocated memory.
		break;
	}
}

void var::copy_memory(const var& value) {
	current_type = value.current_type;
	switch (current_type) {
	case s:	
		s_ptr = new string(*(value.s_ptr));
		break;
	case v:
		v_ptr = new vector<var>(*(value.v_ptr));
		break;
	case m:
		m_ptr = new map<string, var>(*(value.m_ptr));
		break;
	default:
		memcpy(this, &value, sizeof(var));
		break;
	}
}

string var::cast_current_to_string() const {
	ostringstream buffer;

	switch (current_type) {
	case e:
		break;
	case ui:
		buffer << ui_value;
		break;
	case uli:
		buffer << uli_value;
		break;
	case ulli:
		buffer << ulli_value;
		break;
	case i:
		buffer << i_value;
		break;
	case li:
		buffer << li_value;
		break;
	case lli:
		buffer << lli_value;
		break;
	case f:
		buffer << f_value;
		break;
	case d:
		buffer << d_value;
		break;
	case s:
		return *s_ptr;
		break;
	default:
		throw_conversion_exception();
	}

	return buffer.str();
}

template <class T>
T var::cast_string_to_other(string str) {
	istringstream input;
	T result;
	string temp;

	chop(str);
	input.str(str);
	input.exceptions(istringstream::failbit);

	try {
		input >> result;
	} catch (exception& e) {
		throw_conversion_exception();
	}

	// Make sure there is nothing uninterpreted in buffer.
	// Exception expected.
	try {
		input >> temp;
	} catch (exception& e) {
		return result;
	}

	throw_conversion_exception();
	return result; // just to mute compiler warning.
}

template <class T>
T var::cast_current_to_other() const {
	return cast_string_to_other<T>(cast_current_to_string());
}

bool var::is_space(char x) {
	return isspace(x);
}

void var::chop(string& str) {
	string::iterator itr;
	string::reverse_iterator back_itr;

	for (back_itr = str.rbegin(); back_itr != str.rend(); ++back_itr)
		if (!is_space(*back_itr))
			break;
	str.erase(str.length() - (back_itr - str.rbegin()));

	for (itr = str.begin(); itr != str.end(); ++itr)
		if (!is_space(*itr))
			break;
	str.erase(str.begin(), itr);
}

void var::throw_conversion_exception() {
	var_conversion_exception e;
	throw e;
}

void var::throw_bound_exception() {
	var_bound_exception e;
	throw e;
}

var::var() {
	current_type = e;
}

var::var(const var& value) {
	copy_memory(value);
}

var& var::operator = (const var& value) {
	// Avoid self-copying. This is meaningless and dangerous.
	if (this == &value)
		return *this;

	free_memory();
	copy_memory(value);
	return *this;
}

var::var(unsigned int value) {
	current_type = ui;
	ui_value = value;
}

var::var(unsigned long int value) {
	current_type = uli;
	uli_value = value;
}

var::var(unsigned long long int value) {
	current_type = ulli;
	ulli_value = value;
}

var::var(int value) {
	current_type = i;
	i_value = value;
}

var::var(long int value) {
	current_type = li;
	li_value = value;
}

var::var(long long int value) {
	current_type = lli;
	lli_value = value;
}

var::var(float value) {
	current_type = f;
	f_value = value;
}

var::var(double value) {
	current_type = d;
	d_value = value;
}

var::var(const char* value) {
	current_type = s;
	s_ptr = new string(value);
}

var::var(const string& value) {
	current_type = s;
	s_ptr = new string(value);
}

var::~var() {
	free_memory();
}

var& var::operator [] (int subscript) {
	size_t new_subscript;

	// Try to convert the index.
	try {
		new_subscript = (size_t)(var)subscript;
	} catch (const exception& e) {
		throw_bound_exception();
	}

	return (*this)[new_subscript];
}

var& var::operator [] (size_t subscript) {
	if (current_type == e) {
		current_type = v;
		v_ptr = new vector<var>();
	}

	if (current_type != v)
		throw_conversion_exception();

	// Check the bounding. We accept subscript == v_ptr->size() in case
	// the user wish to insert something new.
	if (subscript < 0 || subscript > v_ptr->size())
		throw_bound_exception();

	// If the index is just larger, the user means to push something new.
	if (subscript == v_ptr->size())
		v_ptr->push_back(var());

	return (*v_ptr)[subscript];
}

const var& var::operator [] (int subscript) const {
	size_t new_subscript;

	// Try to convert the index.
	try {
		new_subscript = (size_t)(var)subscript;
	} catch (const exception& e) {
		throw_bound_exception();
	}

	return (*this)[new_subscript];
}

const var& var::operator [] (size_t subscript) const {
	if (current_type != v)
		throw_conversion_exception();

	// Unlike the non-const function, we do not allow pushing new things.
	// Therefore we ban having subscript == vptr->size().
	if (subscript < 0 || subscript >= v_ptr->size())
		throw_bound_exception();

	return (*v_ptr)[subscript];
}

void var::push_back(const var& value) {
	if (current_type == e) {
		current_type = v;
		v_ptr = new vector<var>();
	}

	if (current_type != v)
		throw_conversion_exception();

	v_ptr->push_back(value);
}

void var::pop_back() {
	if (current_type == e) {
		current_type = v;
		v_ptr = new vector<var>();
	}

	if (current_type != v)
		throw_conversion_exception();

	if (v_ptr->size())
		v_ptr->pop_back();
}

var& var::operator [] (const char* subscript) {
	string temp(subscript);
	return (*this)[temp];
}

var& var::operator [] (const string& subscript) {
	if (current_type == e) {
		current_type = m;
		m_ptr = new map<string, var>();
	}

	if (current_type != m)
		throw_conversion_exception();

	return (*m_ptr)[subscript];
}

const var::var& var::operator [] (const char* subscript) const {
	string temp(subscript);
	return (const var&)((*this)[temp]);
}

const var::var& var::operator [] (const string& subscript) const {
	if (current_type != m)
		throw_conversion_exception();

	// We can't use map<...>[...] directly since this would immediately
	// create an empty entry, and interfere the const-ness.
	map<string, var>::iterator result = m_ptr->find(subscript);
	if (result == m_ptr->end())
		throw_bound_exception();

	return result->second;
}

var var::get_keys() const {
	var result;

	if (current_type == e)
		return result;

	if (current_type != m)
		throw_conversion_exception();

	for (map<string, var>::iterator itr = m_ptr->begin();
	     itr != m_ptr->end(); ++itr)
		result[result.size()] = itr->first;

	return result;
}

void var::erase(const char* subscript) {
	string temp(subscript);
	erase(temp);
}

void var::erase(const string& subscript) {
	if (current_type == e) {
		current_type = m;
		m_ptr = new map<string, var>();
	}

	if (current_type != m)
		throw_conversion_exception();

	m_ptr->erase(subscript);
}

size_t var::size() const {
	switch (current_type) {
	case e:
		return 0;
		break;
	case v:
		return v_ptr->size();
		break;
	case m:
		return m_ptr->size();
		break;
	default:
		throw_conversion_exception();
		break;
	}

	return 0; // just to mute the compiler warning.
}

bool var::is_empty() const {
	return (current_type == e);
}

bool var::is_vector() const {
	return (current_type == v);
}

bool var::is_map() const {
	return (current_type == m);
}

var::operator unsigned int() const {
	if (current_type == ui)
		return ui_value;
	else
		return cast_current_to_other<unsigned int>();
}

var::operator unsigned long int() const {
	if (current_type == uli)
		return uli_value;
	else
		return cast_current_to_other<unsigned long int>();
}

var::operator unsigned long long int() const {
	if (current_type == ulli)
		return ulli_value;
	else
		return cast_current_to_other<unsigned long long int>();
}

var::operator int() const {
	if (current_type == i)
		return i_value;
	else
		return cast_current_to_other<int>();
}

var::operator long int() const {
	if (current_type == li)
		return li_value;
	else
		return cast_current_to_other<long int>();
}

var::operator long long int() const {
	if (current_type == lli)
		return lli_value;
	else
		return cast_current_to_other<long long int>();
}

var::operator float() const {
	if (current_type == f)
		return f_value;
	else
		return cast_current_to_other<float>();
}

var::operator double() const {
	if (current_type == d)
		return d_value;
	else
		return cast_current_to_other<double>();
}

var::operator const char*() const {
	static string cache;
	if (current_type == s) {
		return s_ptr->c_str();
	} else {
		cache = cast_current_to_string();
		return cache.c_str();
	}
}

/*
var::operator string() const {
	if (current_type == s)
		return *s_ptr;
	else
		return cast_current_to_string();
}
*/

/*
string& operator = (string& to_assign, const var& value) {
	if (value.current_type == value.s)
		to_assign = *value.s_ptr;
	else
		to_assign = value.cast_current_to_string();
	return to_assign;
}
*/

const char* var_failure::what() const throw() {
	return "Variable failure.";
}

const char* var_conversion_exception::what() const throw() {
	return "Invalid conversion.";
}

const char* var_bound_exception::what() const throw() {
	return "Invalid subscript.";
}

istream& operator >> (istream& stream, var& value) {
	string buffer;
	stream >> buffer;
	value = buffer;
	return stream;
}

ostream& operator << (ostream& stream, const var& value) {
	stream << (string)value;
	return stream;
}

bool getline(istream& stream, var& value) {
	string buffer;
	if (getline(stream, buffer)) {
		value = buffer;
		return true;
	} else {
		return false;
	}
}

} // end of namespace sxx_var.

