#ifndef _SXX_VAR_
#define _SXX_VAR_

#include <string>
#include <vector>
#include <map>
#include <exception>
using std::istream;
using std::ostream;
using std::string;
using std::vector;
using std::map;
using std::exception;

namespace sxx_var {

class var {
private:
	static const int e = 0; // empty.
	static const int ui = 0x1; // unsigned int.
	static const int uli = 0x2; // unsinged long int.
	static const int ulli = 0x4; // unsigned long long int.

	static const int i = 0x11; // int.
	static const int li = 0x12; // long int.
	static const int lli = 0x14; // long long int.

	static const int f = 0x24; // float.
	static const int d = 0x28; // double.

	static const int s = 0xff; // string.

	static const int v = 0x100; // vector. size_t -> var.
	static const int m = 0x200; // map. string -> var.

	int current_type;
	union {
		unsigned int ui_value;
		unsigned long int uli_value;
		unsigned long long int ulli_value;

		int i_value;
		long int li_value;
		long long int lli_value;

		float f_value;
		double d_value;

		string* s_ptr;

		vector<var>* v_ptr;
		map<string, var>* m_ptr;
	};

	void free_memory();
	void copy_memory(const var& value);

	string cast_current_to_string() const;
	template <class T> static T cast_string_to_other(string str);
	template <class T> T cast_current_to_other() const;

	static bool is_space(char x);
	static void chop(string& str);

	static void throw_conversion_exception();
	static void throw_bound_exception();

public:
	var();
	var(const var& value);
	var& operator = (const var& value);

	var(unsigned int value);
	var(unsigned long int value);
	var(unsigned long long int value);

	var(int value);
	var(long int value);
	var(long long int value);

	var(float value);
	var(double value);

	var(const char* value);
	var(const string& value);

	~var();

	// Operators for vector.
	var& operator [] (int subscript);
	var& operator [] (size_t subscript);
	const var& operator [] (int subscript) const;
	const var& operator [] (size_t subscript) const;
	void push_back(const var& value);
	void pop_back();

	// Operations for map.
	var& operator [] (const char* subscript);
	var& operator [] (const string& subscript);
	const var& operator [] (const char* subscript) const;
	const var& operator [] (const string& subscript) const;
	var get_keys() const;
	void erase(const char* subscript);
	void erase(const string& subscript);

	// For both vector and map.
	bool is_empty() const;
	bool is_vector() const;
	bool is_map() const;
	size_t size() const;

	operator unsigned int() const;
	operator unsigned long int() const;
	operator unsigned long long int() const;

	operator int() const;
	operator long int() const;
	operator long long int() const;

	operator float() const;
	operator double() const;

	operator const char*() const;
	// operator string() const;

	friend istream& operator >> (istream& stream, var& value);
	friend ostream& operator << (ostream& stream, const var& value);
	friend bool getline(istream& stream, var& value);

};

class var_failure: public exception {

public:
	virtual const char* what() const throw();

};

class var_conversion_exception: public var_failure {

public:
	virtual const char* what() const throw();

};

class var_bound_exception: public var_failure {

public:
	virtual const char* what() const throw();

};

} // end of namespace sxx_var.

#endif

